/* eslint-disable @typescript-eslint/no-var-requires */
// 这里webpack loader 使用 commonjs 语法
const path = require('path')
const fs = require('fs-extra')
const babel = require('@babel/core')

/**
 * 在入口代码末尾增加导出类代码
 * 只处理入口文件，减少复杂度，可能在其它文件里也有和入口同名的 class
 * @param {string} source
 * @returns {string}
 */
module.exports = function (source) {
  const allExportClasses = new Set()
  let entryParsed = false // 入口类是否已经解析

  const ast = babel.parseSync(source, {
    root: path.resolve(__dirname, '.'),
  })

  let pkgJson = null

  const pkgJsonFile = path.join(path.dirname(this.resourcePath), 'bundle.json')
  if (fs.existsSync(pkgJsonFile)) {
    pkgJson = require(pkgJsonFile)
    pkgJson.name = pkgJson.name || 'plugin'
  }

  babel.traverse(ast, {
    // 处理 export default 的 class
    ExportDefaultDeclaration(nodePath) {
      collectClassNames(nodePath)
    },
    // 处理 export 的 class
    ExportNamedDeclaration(nodePath) {
      collectClassNames(nodePath)
    },
    // 处理入口类，如果没有写 export 也能导出
    ClassDeclaration(nodePath) {
      collectEntryClassName(nodePath)
    },
  })

  function collectEntryClassName(nodePath) {
    if (
      !entryParsed &&
      nodePath.node.superClass?.object?.name === 'THING' &&
      nodePath.node.superClass?.property?.name === 'BasePlugin' &&
      nodePath.node.type === 'ClassDeclaration'
    ) {
      const name = nodePath.node.id.name

      // 实例名取值： bundle.json 的 name -> 类名转驼峰
      allExportClasses.add([pkgJson.name, name])
      entryParsed = true
    }
  }

  function collectClassNames(nodePath) {
    // THING.BasePlugin 类
    if (nodePath.node.declaration?.superClass) {
      collectEntryClassName({ node: nodePath.node.declaration })
    } else if (nodePath.node.declaration?.id?.name) {
      // 处理export class A, export default class A
      const name = nodePath.node.declaration.id.name
      allExportClasses.add([name, name])
    } else if (nodePath.node.specifiers?.length) {
      // 处理 export { A, B }, export * as B from
      nodePath.node.specifiers.forEach((s) => {
        const name = s.exported.name
        allExportClasses.add([name, name])
      })
    } else if (nodePath.node.declaration?.properties?.length) {
      // 处理 export default { A, B }
      nodePath.node.declaration.properties.forEach((s) => {
        const name = s.key.name
        allExportClasses.add([name, name])
      })
    } else if (nodePath.node.declaration?.name) {
      // 处理 export default A
      const name = nodePath.node.declaration.name
      allExportClasses.add([name, name])
    }
  }

  // 在代码末尾增加 THING.EXTEND.Utils.exportPluginClasses({A:A, B:B})
  return (
    source +
    [
      `THING.EXTEND.Utils.exportPluginClasses({`,
      Array.from(allExportClasses)
        .map((c) => {
          return `${c[0]}: ${c[1]}`
        })
        .join(','),
      `})`,
    ].join('')
  )
}
